/**************************************************************************************************
 * This contains templates for converting various data-types to a char[].
 *
 * Copyright (c) 2007 Diggory Hardy.
 * Licensed under the Academic Free License version 3.0
 *
 * This module basically implements the following templated function for $(B most) basic D types:
 * bool, byte, short, int, long, ubyte, ushort, uint, ulong, float, double, real, char.
 * It also supports arrays of any supported type (including of other arrays) and has special
 * handling for strings (char[]) and binary (ubyte[]) data-types.
 * -----------------------------
 * char[] format(T) (T value);
 * -----------------------------
 *
 * There are also a few utility functions defined; the public ones have their own documentation.
 *
 * On errors, a warning is logged and an MTConvertException is thrown. No other exceptions should
 * be thrown and none thrown from functions used outside this module.
 *************************************************************************************************/
module mde.mergetag.format;

// package imports
import mde.mergetag.exception;

// tango imports
import cInt = tango.text.convert.Integer;
import cFloat = tango.text.convert.Float;
import Utf = tango.text.convert.Utf;
import Util = tango.text.Util;
import tango.util.log.Log : Log, Logger;

private Logger logger;
static this () {
    logger = Log.getLogger ("mde.mergetag.format");
}

//BEGIN Convert templates
/* Idea: could extend format with a second parameter, containing flags for things like base to output.
 * Unnecessary for mergetag though.
 */
// Arrays
char[] format(T : T[]) (T[] val) {
    char[val.length * defLength!(T)] ret = void;
    ret[0] = '[';
    uint i = 0;
    foreach (T x; val) {
        char[] s = format!(T) (x);
        i += s.length;
        if (i >= ret.length) ret.length = ret.length * 2;	// check.
        ret[i-s.length .. i] = s;
        ret[i] = ',';
    }
    ret[i++] = ']';	// replaces last comma
    return ret[0..i];
}
char[] format(T : dchar[]) (T val) {
    return format (toUtf8 (val));
}
char[] format(T : wchar[]) (T val) {
    return format (toUtf8 (val));
}
char[] format(T : char[]) (T val) {
    char[val.length * 2 + 2] ret = void;	// Initial storage. This should ALWAYS be enough.
    ret[0] = '"';
    uint i = 0;
    for (uint t = 0; t < val.length;) {
        // process a block of non-escapable characters
        uint s = t;
        while (t < val.length && !isEscapableChar(val[t]))
            ++t;	// skip all non-escapable chars
        uint j = i + t - s;
        ret[i..j] = [s..t];	// copy a block
        i = j;
        // process a block of escapable charaters
        while (t < val.length && isEscapableChar(val[t])) {
            ret[i++] = '\\';				// backslash; increment i
            ret[i++] = replaceEscapableChar(val[t++]);	// character; increment i and t
        }
    }
    ret[i++] = '"';
    return ret[0..i];
}
char[] format(T : ubyte[]) (T val) {
    static const char[16] digits = "0123456789abcdef";
    
    char[val.length * 2] ret = void;	// exact length
    uint i = 0;
    foreach (ubyte x; val) {
        ret[i++] = digits[x >> 4];
        ret[i++] = digits[x & 0x0F];
    }
    return ret;
}

// Support for outputting a wide char... I reccomend against trying to output these though.
const char[] WIDE_CHAR_ERROR = "Error: unicode character cannot be converted to a single UTF-8 char";
char[] format(T : dchar) (T val) {
    if (val <= 127u) return format (cast(char) val);	// this char can be converted
    throwMTFException (WIDE_CHAR_ERROR);
}
char[] format(T : wchar) (T val) {
    if (val <= 127u) return format (cast(char) val);	// this char can be converted
    throwMTFException (WIDE_CHAR_ERROR);
}
char[] format(T : char) (T val) {
    // Note: if (val > 127) "is invalid UTF-8 single char"
    // However we don't know what this is for, in particular if it will be recombined with other chars later
    
    char[4] ret;	// max length for an escaped char
// FIXME: carry on down
    if (src.length < 3 || src[0] != '\'' || src[$-1] != '\'')
        throwMTCException ("Invalid char: not quoted (\'*\')");
    src = src[1..$-1];
    uint pos;
    char ret = toChar (src, pos);
    if (pos < src.length) {
        if (ret & 0xC0u) throwMTCException ("Invalid char: too long (non-ASCII UTF-8 characters cannot be read as a single character)");
        else throwMTCException ("Invalid char: too long");
    }
    return ret;
}

T format(T : bool) (char[] src) {
    src = Util.trim(src);
    if (src == "true") return true;
    if (src == "false") return false;
    uint pos;
    while (src.length > pos && src[pos] == '0') ++pos;	// strip leading zeros
    if (src.length == pos && pos > 0) return false;
    if (src.length == pos + 1 && src[pos] == '1') return true;
    throwMTCException ("Invalid bool: not true or false and doesn't evaluate to 0 or 1");
}
T format(T : byte) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : short) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : int) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : long) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : ubyte) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : ushort) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : uint) (char[] src) {
    return toTInt!(T) (src);
}
T format(T : ulong) (char[] src) {
    return toTInt!(T) (src);
}

T format(T : float) (char[] src) {
    return toTFloat!(T) (src);
}
T format(T : double) (char[] src) {
    return toTFloat!(T) (src);
}
T format(T : real) (char[] src) {
    return toTFloat!(T) (src);
}
//END Convert templates

//BEGIN Length templates
/* This template provides the initial length for strings for formatting various types. These strings
 * can be expanded; this value should cover 90% of cases or so.
 * FIXME: provide more specialisations
 */
private {
    template defLength(T) {	const uint defLength = 20;	}
}
//END Length templates

//BEGIN Utility funcs
/** Templated read-int function to read (un)signed 1-4 byte integers.
 *
 * Actually a reimplementation of tango.text.convert.Integer toLong and parse functions.
 */
TInt toTInt(TInt) (char[] src) {
    const char[] INT_OUT_OF_RANGE = "mde.mergetag.format.toTInt: integer out of range";
    bool sign;
    uint radix, ate, ate2;
    
    ate = cInt.trim (src, sign, radix);
    ulong val = cInt.convert (src[ate..$], radix, &ate2);
    ate += ate2;
    
    while (ate < src.length) {
        if (src[ate] == ' ' || src[ate] == '\t') ++ate;
        else throwMTCException ("mde.mergetag.format.toTInt: invalid integer");
    }
    
    if (val > TInt.max) throwMTCException (INT_OUT_OF_RANGE);
    if (sign) {
        long sval = cast(long) -val;
        if (sval > TInt.min) return cast(TInt) sval;
        else throwMTCException (INT_OUT_OF_RANGE);
    }
    return cast(TInt) val;
}

/** Basically a reimplementation of tango.text.convert.Float.toFloat which checks for trailing
 * whitespace before throwing an exception for overlong input and throws my exception class
 * when it does.
 */
TFloat toTFloat(TFloat) (char[] src) {
    uint ate;

    TFloat x = cFloat.parse (src, &ate);
    while (ate < src.length) {
        if (src[ate] == ' ' || src[ate] == '\t') ++ate;
        else throwMTCException ("mde.mergetag.format.toTFloat: invalid number");
    }
    return x;
}

/** Read a character from a string, with support for escape sequences.
 *
 * Assumes src.length > pos. At return pos is set to one after the last character eaten.
 *
 * Throws an exception on invalid escape sequences. Supported escape sequences are the following
 * subset of those supported by D: \" \' \\ \a \b \f \n \r \t \v
 */
char toChar (char[] src, inout uint pos)
{
    //if (src.length <= pos) throwMTCException ("Invalid char: no character"); // shouldn't happen
    if (src[pos] == '\\') {
        // map of all supported escape sequences
        char[char] escChars;
        escChars['"'] = '"';
        escChars['\''] = '\'';
        escChars['\\'] = '\\';
        escChars['a'] = '\a';
        escChars['b'] = '\b';
        escChars['f'] = '\f';
        escChars['n'] = '\n';
        escChars['r'] = '\r';
        escChars['t'] = '\t';
        escChars['v'] = '\v';
        
        ++pos;
        if (src.length > pos) {
            char* r = src[pos] in escChars;
            ++pos;
            if (r != null) return *r;
        }
        throwMTCException ("Invalid escape sequence.");	// we didn't return, so something failed
    }
    char c = src[pos];
    ++pos;
    return c;
    
    // note on UTF-8 non-ascii chars: these consist of multiple "char"s; can only return one at a
    // time like this anyway. If this is used to read a string it should handle them fine.
}

// Reads one hex char: [0-9A-Fa-f]. Otherwise throws an exception. Doesn't check src.length.
private ubyte readHexChar (char[] src, inout uint pos) {
    ubyte x;
    if (src[pos] >= '0' && src[pos] <= '9') x = src[pos] - '0';
    else if (src[pos] >= 'A' && src[pos] <= 'F') x = src[pos] - 'A' + 10;
    else if (src[pos] >= 'a' && src[pos] <= 'f') x = src[pos] - 'a' + 10;
    else throwMTCException ("Invalid hex digit.");
    ++pos;
    return x;
}

// Generic array reader
private T[] toArray(T : T[]) (char[] src) {
    T[] ret = new T[16];	// avoid unnecessary allocations
    uint i = 0;
    foreach (char[] element; Util.quotes (src[1..$-1],",")) {
        if (i == ret.length) ret.length = ret.length * 2;
        ret[i] = format!(T) (element);
        ++i;
    }
    return ret[0..i];
}
// FIXME: to here.

private bool isEscapableChar (char c) {
    return ((c <= '\r' && c >= '\a') || c == '\"' || c == '\'' || c == '\\')
}
// Warning: this DOES NOT check c is escapable
private char replaceEscapableChar (char c) {
    static char[char] escCharsRev;	// reversed escChars
    static bool escCharsRevFilled;	// will be initialised false
    
    if (!escCharsRevFilled) {	// only do this once
        // map of all supported escape sequences
        escCharsRev['"'] = '"';
        escCharsRev['\''] = '\'';
        escCharsRev['\\'] = '\\';
        escCharsRev['\a'] = 'a';
        escCharsRev['\b'] = 'b';
        escCharsRev['\f'] = 'f';
        escCharsRev['\n'] = 'n';
        escCharsRev['\r'] = 'r';
        escCharsRev['\t'] = 't';
        escCharsRev['\v'] = 'v';
        escCharsRevFilled = true;
    }
    
    return escCharsRev[c];
}

private void throwMTFException (char[] msg) {
    logger.warn (msg);			// only small errors are trapped here
    throw new MTFormatException ();
}
//END Utility funcs
